import json
import unittest
from base64 import b64encode
from textwrap import dedent
from unittest.mock import ANY

from django.test import tag

from anymail.exceptions import AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.postal import PostalInboundWebhookView

from .utils import sample_email_content, sample_image_content
from .utils_postal import ClientWithPostalSignature, make_key
from .webhook_cases import WebhookTestCase


@tag("postal")
@unittest.skipUnless(
    ClientWithPostalSignature, "Install 'cryptography' to run postal webhook tests"
)
class PostalInboundTestCase(WebhookTestCase):
    client_class = ClientWithPostalSignature

    def setUp(self):
        super().setUp()
        self.clear_basic_auth()

        self.client.set_private_key(make_key())

    def test_inbound_basics(self):
        raw_event = {
            "id": 233980,
            "rcpt_to": "test@inbound.example.com",
            "mail_from": "envelope-from@example.org",
            "message": b64encode(
                dedent(
                    """\
                    Received: from mail.example.org by postal.example.com ...
                    Received: by mail.example.org for <test@inbound.example.com> ...
                    DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed; d=example.org; ...
                    MIME-Version: 1.0
                    Received: by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)
                    From: "Displayed From" <from+test@example.org>
                    Date: Wed, 11 Oct 2017 18:31:04 -0700
                    Message-ID: <CAEPk3R+4Zr@mail.example.org>
                    Subject: Test subject
                    To: "Test Inbound" <test@inbound.example.com>, other@example.com
                    Cc: cc@example.com
                    Content-Type: multipart/alternative; boundary="94eb2c05e174adb140055b6339c5"

                    --94eb2c05e174adb140055b6339c5
                    Content-Type: text/plain; charset="UTF-8"
                    Content-Transfer-Encoding: quoted-printable

                    It's a body=E2=80=A6

                    --94eb2c05e174adb140055b6339c5
                    Content-Type: text/html; charset="UTF-8"
                    Content-Transfer-Encoding: quoted-printable

                    <div dir=3D"ltr">It's a body=E2=80=A6</div>

                    --94eb2c05e174adb140055b6339c5--
                    """  # NOQA: E501
                ).encode("utf-8")
            ).decode("ascii"),
            "base64": True,
        }

        response = self.client.post(
            "/anymail/postal/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostalInboundWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        # Postal doesn't provide inbound event timestamp:
        self.assertIsNone(event.timestamp)
        self.assertEqual(event.event_id, 233980)
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\n""",
        )

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertIsNone(message.stripped_text)
        self.assertIsNone(message.stripped_html)
        self.assertIsNone(message.spam_detected)
        self.assertIsNone(message.spam_score)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by postal.example.com ...",
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

    def test_attachments(self):
        image_content = sample_image_content()
        email_content = sample_email_content()
        raw_mime = dedent(
            """\
            MIME-Version: 1.0
            From: from@example.org
            Subject: Attachments
            To: test@inbound.example.com
            Content-Type: multipart/mixed; boundary="boundary0"

            --boundary0
            Content-Type: multipart/related; boundary="boundary1"

            --boundary1
            Content-Type: text/html; charset="UTF-8"

            <div>This is the HTML body. It has an inline image: <img src="cid:abc123">.</div>

            --boundary1
            Content-Type: image/png
            Content-Disposition: inline; filename="image.png"
            Content-ID: <abc123>
            Content-Transfer-Encoding: base64

            {image_content_base64}
            --boundary1--
            --boundary0
            Content-Type: text/plain; charset="UTF-8"
            Content-Disposition: attachment; filename="test.txt"

            test attachment
            --boundary0
            Content-Type: message/rfc822; charset="US-ASCII"
            Content-Disposition: attachment
            X-Comment: (the only valid transfer encodings for message/* are 7bit, 8bit, and binary)

            {email_content}
            --boundary0--
            """  # NOQA: E501
        ).format(
            image_content_base64=b64encode(image_content).decode("ascii"),
            email_content=email_content.decode("ascii"),
        )

        raw_event = {
            "id": 233980,
            "rcpt_to": "test@inbound.example.com",
            "mail_from": "envelope-from@example.org",
            "message": b64encode(raw_mime.encode("utf-8")).decode("ascii"),
            "base64": True,
        }

        response = self.client.post(
            "/anymail/postal/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostalInboundWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "test attachment")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)

    def test_misconfigured_tracking(self):
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set Postal's *tracking* webhook"
            " to Anymail's Postal *inbound* webhook URL.",
        ):
            self.client.post(
                "/anymail/postal/inbound/",
                content_type="application/json",
                data=json.dumps({"status": "Held"}),
            )



import os
import unittest
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin

ANYMAIL_TEST_POSTAL_API_KEY = os.getenv("ANYMAIL_TEST_POSTAL_API_KEY")
ANYMAIL_TEST_POSTAL_API_URL = os.getenv("ANYMAIL_TEST_POSTAL_API_URL")
ANYMAIL_TEST_POSTAL_DOMAIN = os.getenv("ANYMAIL_TEST_POSTAL_DOMAIN")


@tag("postal", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_POSTAL_API_KEY
    and ANYMAIL_TEST_POSTAL_API_URL
    and ANYMAIL_TEST_POSTAL_DOMAIN,
    "Set ANYMAIL_TEST_POSTAL_API_KEY and ANYMAIL_TEST_POSTAL_API_URL and"
    " ANYMAIL_TEST_POSTAL_DOMAIN environment variables to run Postal integration tests",
)
@override_settings(
    ANYMAIL_POSTAL_API_KEY=ANYMAIL_TEST_POSTAL_API_KEY,
    ANYMAIL_POSTAL_API_URL=ANYMAIL_TEST_POSTAL_API_URL,
    EMAIL_BACKEND="anymail.backends.postal.EmailBackend",
)
class PostalBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Postal API integration tests

    These tests run against the **live** Postal API, using the
    environment variable `ANYMAIL_TEST_POSTAL_API_KEY` as the API key and
    `ANYMAIL_TEST_POSTAL_API_URL` as server url.
    If these variables are not set, these tests won't run.
    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_POSTAL_DOMAIN
        self.message = AnymailMessage(
            "Anymail Postal integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Postal send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")
        self.assertGreater(len(message_id), 0)  # non-empty string
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Postal all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            envelope_sender="bounces@%s" % ANYMAIL_TEST_POSTAL_DOMAIN,
            to=["test+to1@anymail.dev", "Recipient 2 <test+to2@anymail.dev>"],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            reply_to=["reply1@example.com"],
            headers={"X-Anymail-Test": "value"},
            tags=["tag 1"],  # max one tag
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")

        message.send()
        self.assertEqual(message.anymail_status.status, {"queued"})
        self.assertEqual(
            message.anymail_status.recipients["test+to1@anymail.dev"].status, "queued"
        )
        self.assertEqual(
            message.anymail_status.recipients["test+to2@anymail.dev"].status, "queued"
        )
        # distinct messages should have different message_ids:
        self.assertNotEqual(
            message.anymail_status.recipients["test+to1@anymail.dev"].message_id,
            message.anymail_status.recipients["teset+to2@anymail.dev"].message_id,
        )

    def test_invalid_from(self):
        self.message.from_email = "webmaster@localhost"  # Django's default From
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        response = err.response.json()
        self.assertEqual(err.status_code, 200)
        self.assertEqual(response["status"], "error")
        self.assertIn(
            "The From address is not authorised to send mail from this server",
            response["data"]["message"],
        )
        self.assertIn("UnauthenticatedFromAddress", response["data"]["code"])

    @override_settings(ANYMAIL_POSTAL_API_KEY="Hey, that's not an API key!")
    def test_invalid_server_token(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        response = err.response.json()
        self.assertEqual(err.status_code, 200)
        self.assertEqual(response["status"], "error")
        self.assertIn(
            "The API token provided in X-Server-API-Key was not valid.",
            response["data"]["message"],
        )
        self.assertIn("InvalidServerAPIKey", response["data"]["code"])



import json
import warnings
from datetime import datetime, timezone
from unittest.mock import ANY, patch

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailConfigurationError, AnymailInsecureWebhookWarning
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.amazon_ses import AmazonSESTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase


class AmazonSESWebhookTestsMixin(SimpleTestCase):
    def post_from_sns(self, path, raw_sns_message, **kwargs):
        return self.client.post(
            path,
            content_type="text/plain; charset=UTF-8",  # SNS posts JSON as text/plain
            data=json.dumps(raw_sns_message),
            HTTP_X_AMZ_SNS_MESSAGE_ID=raw_sns_message["MessageId"],
            HTTP_X_AMZ_SNS_MESSAGE_TYPE=raw_sns_message["Type"],
            # Anymail doesn't use other x-amz-sns-* headers
            **kwargs,
        )


@tag("amazon_ses")
class AmazonSESWebhookSecurityTests(
    AmazonSESWebhookTestsMixin, WebhookBasicAuthTestCase
):
    def call_webhook(self):
        return self.post_from_sns(
            "/anymail/amazon_ses/tracking/",
            {"Type": "Notification", "MessageId": "123", "Message": "{}"},
        )

    # Most actual tests are in WebhookBasicAuthTestCase

    def test_verifies_missing_auth(self):
        # Must handle missing auth header slightly differently from Anymail default 400
        # SuspiciousOperation: SNS will only send basic auth after missing auth responds
        # 401 WWW-Authenticate: Basic realm="..."
        self.clear_basic_auth()
        response = self.call_webhook()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response["WWW-Authenticate"], 'Basic realm="Anymail WEBHOOK_SECRET"'
        )


@tag("amazon_ses")
class AmazonSESNotificationsTests(WebhookTestCase, AmazonSESWebhookTestsMixin):
    def test_bounce_event(self):
        # This test includes a complete Amazon SES example event.
        # (Later tests omit some payload for brevity.)
        # https://docs.aws.amazon.com/ses/latest/DeveloperGuide/notification-examples.html#notification-examples-bounce
        raw_ses_event = {
            "notificationType": "Bounce",
            "bounce": {
                "bounceType": "Permanent",
                "reportingMTA": "dns; email.example.com",
                "bouncedRecipients": [
                    {
                        "emailAddress": "jane@example.com",
                        "status": "5.1.1",
                        "action": "failed",
                        "diagnosticCode": "smtp; 550 5.1.1 <jane@example.com>..."
                        " User unknown",
                    }
                ],
                "bounceSubType": "General",
                # when bounce sent (by receiving ISP):
                "timestamp": "2016-01-27T14:59:44.101Z",
                # unique id for bounce:
                "feedbackId": "00000138111222aa-44455566-cccc"
                "-cccc-cccc-ddddaaaa068a-000000",
                "remoteMtaIp": "127.0.2.0",
            },
            "mail": {
                "timestamp": "2016-01-27T14:59:38.237Z",  # when message sent
                "source": "john@example.com",
                "sourceArn": "arn:aws:ses:us-west-2:888888888888:identity/example.com",
                "sourceIp": "127.0.3.0",
                "sendingAccountId": "123456789012",
                "messageId": "00000138111222aa-33322211-cccc"
                "-cccc-cccc-ddddaaaa0680-000000",
                "destination": [
                    "jane@example.com",
                    "mary@example.com",
                    "richard@example.com",
                ],
                "headersTruncated": False,
                "headers": [
                    {"name": "From", "value": '"John Doe" <john@example.com>'},
                    {
                        "name": "To",
                        "value": '"Jane Doe" <jane@example.com>,'
                        ' "Mary Doe" <mary@example.com>,'
                        ' "Richard Doe" <richard@example.com>',
                    },
                    {"name": "Message-ID", "value": "custom-message-ID"},
                    {"name": "Subject", "value": "Hello"},
                    {"name": "Content-Type", "value": 'text/plain; charset="UTF-8"'},
                    {"name": "Content-Transfer-Encoding", "value": "base64"},
                    {"name": "Date", "value": "Wed, 27 Jan 2016 14:05:45 +0000"},
                    {"name": "X-Tag", "value": "tag 1"},
                    {"name": "X-Tag", "value": "tag 2"},
                    {"name": "X-Metadata", "value": '{"meta1":"string","meta2":2}'},
                ],
                "commonHeaders": {
                    "from": ["John Doe <john@example.com>"],
                    "date": "Wed, 27 Jan 2016 14:05:45 +0000",
                    "to": [
                        "Jane Doe <jane@example.com>, Mary Doe <mary@example.com>,"
                        " Richard Doe <richard@example.com>"
                    ],
                    "messageId": "custom-message-ID",
                    "subject": "Hello",
                },
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            # unique id for SNS event:
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "TopicArn": "arn:aws:sns:us-east-1:1234567890:SES_Events",
            "Subject": "Amazon SES Email Event Notification",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
            "SignatureVersion": "1",
            "Signature": "EXAMPLE-SIGNATURE==",
            "SigningCertURL": "https://sns.us-east-1.amazonaws.com"
            "/SimpleNotificationService-12345abcde.pem",
            "UnsubscribeURL": "https://sns.us-east-1.amazonaws.com"
            "/?Action=Unsubscribe&SubscriptionArn=arn...",
        }

        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_ses_event)
        # timestamp from SNS:
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 26, 17, 58, 59, microsecond=675000, tzinfo=timezone.utc),
        )
        self.assertEqual(
            event.message_id,
            "00000138111222aa-33322211-cccc-cccc-cccc-ddddaaaa0680-000000",
        )
        self.assertEqual(event.event_id, "19ba9823-d7f2-53c1-860e-cb10e0d13dfc")
        self.assertEqual(event.recipient, "jane@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "Permanent: General")
        self.assertEqual(
            event.mta_response, "smtp; 550 5.1.1 <jane@example.com>... User unknown"
        )
        self.assertEqual(event.tags, ["tag 1", "tag 2"])
        self.assertEqual(event.metadata, {"meta1": "string", "meta2": 2})

    # For brevity, remaining tests omit some event fields that aren't used by Anymail

    def test_multiple_bounce_event(self):
        """Amazon SES notification can cover multiple recipients"""
        raw_ses_event = {
            "notificationType": "Bounce",
            "bounce": {
                "bounceType": "Permanent",
                "bounceSubType": "General",
                "bouncedRecipients": [
                    {"emailAddress": "jane@example.com"},
                    {"emailAddress": "richard@example.com"},
                ],
            },
            "mail": {
                "messageId": "00000137860315fd-34208509-5b74"
                "-41f3-95c5-22c1edc3c924-000000",
                "destination": [
                    "jane@example.com",
                    "mary@example.com",
                    "richard@example.com",
                ],
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)

        # tracking handler should be called twice -- once for each bounced recipient
        # (but not for the third, non-bounced recipient)
        self.assertEqual(self.tracking_handler.call_count, 2)

        _, kwargs = self.tracking_handler.call_args_list[0]
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.recipient, "jane@example.com")
        self.assertEqual(event.description, "Permanent: General")
        self.assertIsNone(event.mta_response)

        _, kwargs = self.tracking_handler.call_args_list[1]
        event = kwargs["event"]
        self.assertEqual(event.esp_event, raw_ses_event)
        self.assertEqual(event.recipient, "richard@example.com")

    def test_complaint_event(self):
        raw_ses_event = {
            "notificationType": "Complaint",
            "complaint": {
                "userAgent": "AnyCompany Feedback Loop (V0.01)",
                "complainedRecipients": [{"emailAddress": "richard@example.com"}],
                "complaintFeedbackType": "abuse",
            },
            "mail": {
                "messageId": "000001378603177f-7a5433e7-8edb"
                "-42ae-af10-f0181f34d6ee-000000",
                "destination": [
                    "jane@example.com",
                    "mary@example.com",
                    "richard@example.com",
                ],
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.recipient, "richard@example.com")
        self.assertEqual(event.reject_reason, "spam")
        self.assertEqual(event.description, "abuse")
        self.assertEqual(event.user_agent, "AnyCompany Feedback Loop (V0.01)")

    def test_delivery_event(self):
        raw_ses_event = {
            "notificationType": "Delivery",
            "mail": {
                "timestamp": "2016-01-27T14:59:38.237Z",
                "messageId": "0000014644fe5ef6-9a483358-9170"
                "-4cb4-a269-f5dcdf415321-000000",
                "destination": [
                    "jane@example.com",
                    "mary@example.com",
                    "richard@example.com",
                ],
            },
            "delivery": {
                "timestamp": "2016-01-27T14:59:38.237Z",
                "recipients": ["jane@example.com"],
                "processingTimeMillis": 546,
                "reportingMTA": "a8-70.smtp-out.amazonses.com",
                "smtpResponse": "250 ok:  Message 64111812 accepted",
                "remoteMtaIp": "127.0.2.0",
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.recipient, "jane@example.com")
        self.assertEqual(event.mta_response, "250 ok:  Message 64111812 accepted")

    def test_send_event(self):
        raw_ses_event = {
            "eventType": "Send",
            "mail": {
                "timestamp": "2016-10-14T05:02:16.645Z",
                "messageId": "7c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000",
                "destination": ["recipient@example.com"],
                "tags": {
                    "ses:configuration-set": ["ConfigSet"],
                    "ses:source-ip": ["192.0.2.0"],
                    "ses:from-domain": ["example.com"],
                    "ses:caller-identity": ["ses_user"],
                    "myCustomTag1": ["myCustomTagValue1"],
                    "myCustomTag2": ["myCustomTagValue2"],
                },
            },
            "send": {},
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "sent")
        self.assertEqual(event.esp_event, raw_ses_event)
        # timestamp from SNS:
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 26, 17, 58, 59, microsecond=675000, tzinfo=timezone.utc),
        )
        self.assertEqual(
            event.message_id, "7c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000"
        )
        self.assertEqual(event.event_id, "19ba9823-d7f2-53c1-860e-cb10e0d13dfc")
        self.assertEqual(event.recipient, "recipient@example.com")
        # Anymail doesn't load Amazon SES "Message Tags":
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_reject_event(self):
        raw_ses_event = {
            "eventType": "Reject",
            "mail": {
                "timestamp": "2016-10-14T17:38:15.211Z",
                "messageId": "7c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000",
                "destination": ["recipient@example.com"],
            },
            "reject": {"reason": "Bad content"},
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.reject_reason, "blocked")
        self.assertEqual(event.description, "Bad content")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_open_event(self):
        raw_ses_event = {
            "eventType": "Open",
            "mail": {
                "destination": ["recipient@example.com"],
                "messageId": "7c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000",
            },
            "open": {
                "ipAddress": "192.0.2.1",
                "timestamp": "2017-08-09T22:00:19.652Z",
                "userAgent": "Mozilla/5.0"
                " (iPhone; CPU iPhone OS 10_3_3 like Mac OS X)...",
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_3 like Mac OS X)...",
        )

    def test_click_event(self):
        raw_ses_event = {
            "eventType": "Click",
            "click": {
                "ipAddress": "192.0.2.1",
                "link": "https://docs.aws.amazon.com/ses/latest/DeveloperGuide/",
                "linkTags": {
                    "samplekey0": ["samplevalue0"],
                    "samplekey1": ["samplevalue1"],
                },
                "timestamp": "2017-08-09T23:51:25.570Z",
                "userAgent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
                " AppleWebKit/537.36...",
            },
            "mail": {
                "destination": ["recipient@example.com"],
                "messageId": "7c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000",
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36...",
        )
        self.assertEqual(
            event.click_url, "https://docs.aws.amazon.com/ses/latest/DeveloperGuide/"
        )

    def test_rendering_failure_event(self):
        raw_ses_event = {
            "eventType": "Rendering Failure",
            "mail": {
                "messageId": "c191be45-e9aedb9a-02f9-4d12-a87d-dd0099a07f8a-000000",
                "destination": ["recipient@example.com"],
            },
            "failure": {
                "errorMessage": "Attribute 'attributeName' is not present"
                " in the rendering data.",
                "templateName": "MyTemplate",
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "19ba9823-d7f2-53c1-860e-cb10e0d13dfc",
            "Message": json.dumps(raw_ses_event) + "\n",
            "Timestamp": "2018-03-26T17:58:59.675Z",
        }
        response = self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=AmazonSESTrackingWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "failed")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.description,
            "Attribute 'attributeName' is not present in the rendering data.",
        )

    def test_incorrect_received_event(self):
        """The tracking webhook should warn if it receives inbound events"""
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
            "Message": '{"notificationType": "Received"}',
        }
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set an Amazon SES *inbound* receipt rule to publish to an"
            " SNS Topic that posts to Anymail's *tracking* webhook URL. (SNS TopicArn"
            " arn:aws:sns:us-east-1:111111111111:SES_Inbound)",
        ):
            self.post_from_sns("/anymail/amazon_ses/tracking/", raw_sns_message)


@tag("amazon_ses")
class AmazonSESSubscriptionManagementTests(WebhookTestCase, AmazonSESWebhookTestsMixin):
    # Anymail will automatically respond to SNS subscription notifications
    # if Anymail is configured to require basic auth via WEBHOOK_SECRET.
    # (Note that WebhookTestCase sets up ANYMAIL WEBHOOK_SECRET.)

    def setUp(self):
        super().setUp()
        # Mock boto3.session.Session().client('sns').confirm_subscription (and any
        # other client operations). (We could also use botocore.stub.Stubber, but mock
        # works well with our test structure.)
        self.patch_boto3_session = patch(
            "anymail.webhooks.amazon_ses.boto3.session.Session", autospec=True
        )
        #: boto3.session.Session
        self.mock_session = self.patch_boto3_session.start()
        self.addCleanup(self.patch_boto3_session.stop)
        #: boto3.session.Session().client
        self.mock_client = self.mock_session.return_value.client
        #: boto3.session.Session().client('sns', ...)
        self.mock_client_instance = self.mock_client.return_value
        self.mock_client_instance.confirm_subscription.return_value = {
            "SubscriptionArn": "arn:aws:sns:us-west-2:123456789012:SES_Notifications"
            ":aaaaaaa-..."
        }

    SNS_SUBSCRIPTION_CONFIRMATION = {
        "Type": "SubscriptionConfirmation",
        "MessageId": "165545c9-2a5c-472c-8df2-7ff2be2b3b1b",
        "Token": "EXAMPLE_TOKEN",
        "TopicArn": "arn:aws:sns:us-west-2:123456789012:SES_Notifications",
        "Message": "You have chosen to subscribe ...\n"
        "To confirm..., visit the SubscribeURL included in this message.",
        "SubscribeURL": "https://sns.us-west-2.amazonaws.com/"
        "?Action=ConfirmSubscription&TopicArn=...",
        "Timestamp": "2012-04-26T20:45:04.751Z",
        "SignatureVersion": "1",
        "Signature": "EXAMPLE-SIGNATURE==",
        "SigningCertURL": "https://sns.us-west-2.amazonaws.com/"
        "SimpleNotificationService-12345abcde.pem",
    }

    def test_sns_subscription_auto_confirmation(self):
        """Anymail webhook will auto-confirm SNS topic subscriptions"""
        response = self.post_from_sns(
            "/anymail/amazon_ses/tracking/", self.SNS_SUBSCRIPTION_CONFIRMATION
        )
        self.assertEqual(response.status_code, 200)
        # auto-confirmed:
        self.mock_client.assert_called_once_with(
            "sns", config=ANY, region_name="us-west-2"
        )
        self.mock_client_instance.confirm_subscription.assert_called_once_with(
            TopicArn="arn:aws:sns:us-west-2:123456789012:SES_Notifications",
            Token="EXAMPLE_TOKEN",
            AuthenticateOnUnsubscribe="true",
        )
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)

    def test_sns_subscription_confirmation_failure(self):
        """Auto-confirmation allows error through if confirm call fails"""
        from botocore.exceptions import ClientError

        self.mock_client_instance.confirm_subscription.side_effect = ClientError(
            {
                "Error": {
                    "Type": "Sender",
                    "Code": "InternalError",
                    "Message": "Gremlins!",
                },
                "ResponseMetadata": {
                    "RequestId": "aaaaaaaa-2222-1111-8888-bbbb3333bbbb",
                    "HTTPStatusCode": 500,
                },
            },
            operation_name="confirm_subscription",
        )
        with self.assertRaisesMessage(ClientError, "Gremlins!"):
            self.post_from_sns(
                "/anymail/amazon_ses/tracking/", self.SNS_SUBSCRIPTION_CONFIRMATION
            )
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)

    @override_settings(ANYMAIL_AMAZON_SES_CLIENT_PARAMS={"region_name": "us-east-1"})
    def test_sns_subscription_confirmation_different_region(self):
        """
        Anymail confirms the subscription in the SNS Topic's own region,
        rather than any default region
        """
        # (The SNS_SUBSCRIPTION_CONFIRMATION above has a TopicArn in region us-west-2)
        self.post_from_sns(
            "/anymail/amazon_ses/tracking/", self.SNS_SUBSCRIPTION_CONFIRMATION
        )
        self.mock_client.assert_called_once_with(
            "sns", config=ANY, region_name="us-west-2"
        )

    # clear WEBHOOK_SECRET setting from base WebhookTestCase
    @override_settings(ANYMAIL={})
    def test_sns_subscription_confirmation_auth_disabled(self):
        """
        Anymail *won't* auto-confirm SNS subscriptions if WEBHOOK_SECRET isn't in use
        """
        # (warning gets tested elsewhere)
        warnings.simplefilter("ignore", AnymailInsecureWebhookWarning)
        with self.assertLogs("django.security.AnymailWebhookValidationFailure") as cm:
            response = self.post_from_sns(
                "/anymail/amazon_ses/tracking/", self.SNS_SUBSCRIPTION_CONFIRMATION
            )
        self.assertEqual(response.status_code, 400)  # bad request
        self.assertEqual(
            [
                "Anymail received an unexpected SubscriptionConfirmation request for"
                " Amazon SNS topic"
                " 'arn:aws:sns:us-west-2:123456789012:SES_Notifications'."
                " (Anymail can automatically confirm SNS subscriptions if you set"
                " a WEBHOOK_SECRET and use that in your SNS notification url."
                " Or you can manually confirm this subscription in the SNS dashboard"
                " with token 'EXAMPLE_TOKEN'.)"
            ],
            [record.getMessage() for record in cm.records],
        )
        # *didn't* try to confirm the subscription:
        self.assertEqual(self.mock_client_instance.confirm_subscription.call_count, 0)
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)

    def test_sns_confirmation_success_notification(self):
        """
        Anymail ignores the 'Successfully validated' notification
        after confirming an SNS subscription
        """
        response = self.post_from_sns(
            "/anymail/amazon_ses/tracking/",
            {
                "Type": "Notification",
                "MessageId": "7fbca0d9-eeab-5285-ae27-f3f57f2e84b0",
                "TopicArn": "arn:aws:sns:us-west-2:123456789012:SES_Notifications",
                "Message": "Successfully validated SNS topic"
                " for Amazon SES event publishing.",
                "Timestamp": "2018-03-21T16:58:45.077Z",
                "SignatureVersion": "1",
                "Signature": "EXAMPLE_SIGNATURE==",
                "SigningCertURL": "https://sns.us-east-1.amazonaws.com"
                "/SimpleNotificationService-12345abcde.pem",
                "UnsubscribeURL": "https://sns.us-east-1.amazonaws.com"
                "/?Action=Unsubscribe...",
            },
        )
        self.assertEqual(response.status_code, 200)
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)

    def test_sns_unsubscribe_confirmation(self):
        """
        Anymail ignores the UnsubscribeConfirmation SNS message
        after deleting a subscription
        """
        response = self.post_from_sns(
            "/anymail/amazon_ses/tracking/",
            {
                "Type": "UnsubscribeConfirmation",
                "MessageId": "47138184-6831-46b8-8f7c-afc488602d7d",
                "Token": "EXAMPLE_TOKEN",
                "TopicArn": "arn:aws:sns:us-west-2:123456789012:SES_Notifications",
                "Message": "You have chosen to deactivate subscription ...\n"
                "To cancel ... visit the SubscribeURL...",
                "SubscribeURL": "https://sns.us-west-2.amazonaws.com"
                "/?Action=ConfirmSubscription&TopicArn=...",
                "Timestamp": "2012-04-26T20:06:41.581Z",
                "SignatureVersion": "1",
                "Signature": "EXAMPLE_SIGNATURE==",
                "SigningCertURL": "https://sns.us-east-1.amazonaws.com"
                "/SimpleNotificationService-12345abcde.pem",
            },
        )
        self.assertEqual(response.status_code, 200)
        # *didn't* try to use the Token to re-enable the subscription:
        self.assertEqual(self.mock_client_instance.confirm_subscription.call_count, 0)
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)

    @override_settings(ANYMAIL_AMAZON_SES_AUTO_CONFIRM_SNS_SUBSCRIPTIONS=False)
    def test_disable_auto_confirmation(self):
        """
        The ANYMAIL setting AMAZON_SES_AUTO_CONFIRM_SNS_SUBSCRIPTIONS
        will disable this feature
        """
        response = self.post_from_sns(
            "/anymail/amazon_ses/tracking/", self.SNS_SUBSCRIPTION_CONFIRMATION
        )
        self.assertEqual(response.status_code, 200)
        # *didn't* try to subscribe:
        self.assertEqual(self.mock_session.call_count, 0)
        self.assertEqual(self.mock_client.call_count, 0)
        # didn't notify receivers:
        self.assertEqual(self.tracking_handler.call_count, 0)
        self.assertEqual(self.inbound_handler.call_count, 0)



import json
from datetime import date, datetime, timezone
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.text import MIMEText

from django.core import mail
from django.test import override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailRecipientsRefused,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage, attach_inline_image_file

from .mock_requests_backend import RequestsBackendMockAPITestCase
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("sparkpost")
@override_settings(
    EMAIL_BACKEND="anymail.backends.sparkpost.EmailBackend",
    ANYMAIL={"SPARKPOST_API_KEY": "test_api_key"},
)
class SparkPostBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    """TestCase that uses SparkPostEmailBackend with a mocked transmissions.send API"""

    DEFAULT_RAW_RESPONSE = b"""{
        "results": {
            "id": "12345678901234567890",
            "total_accepted_recipients": 1,
            "total_rejected_recipients": 0
        }
    }"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    def set_mock_result(self, accepted=1, rejected=0, id="12345678901234567890"):
        """Set a mock response that reflects count of accepted/rejected recipients"""
        raw = json.dumps(
            {
                "results": {
                    "id": id,
                    "total_accepted_recipients": accepted,
                    "total_rejected_recipients": rejected,
                }
            }
        ).encode("utf-8")
        self.set_mock_response(raw=raw)
        return raw


@tag("sparkpost")
class SparkPostBackendStandardEmailTests(SparkPostBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )

        self.assert_esp_called("/api/v1/transmissions/")

        headers = self.get_api_call_headers()
        self.assertEqual("test_api_key", headers["Authorization"])

        data = self.get_api_call_json()
        self.assertEqual(data["content"]["subject"], "Subject here")
        self.assertEqual(data["content"]["text"], "Here is the message.")
        self.assertEqual(data["content"]["from"], "from@example.com")
        self.assertEqual(
            data["recipients"],
            [{"address": {"email": "to@example.com", "header_to": "to@example.com"}}],
        )

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        self.set_mock_result(accepted=6)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()

        data = self.get_api_call_json()
        self.assertEqual(data["content"]["from"], "From Name <from@example.com>")
        # This also checks recipient generation for cc and bcc. Because it's *not* a
        # batch send, each recipient should see a To header reflecting all To addresses.
        self.assertCountEqual(
            data["recipients"],
            [
                {
                    "address": {
                        "email": "to1@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
                {
                    "address": {
                        "email": "to2@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
                # cc and bcc must be explicitly specified as recipients
                {
                    "address": {
                        "email": "cc1@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
                {
                    "address": {
                        "email": "cc2@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
                {
                    "address": {
                        "email": "bcc1@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
                {
                    "address": {
                        "email": "bcc2@example.com",
                        "header_to": "Recipient #1 <to1@example.com>, to2@example.com",
                    }
                },
            ],
        )
        # Make sure we added a formatted Cc header visible to recipients
        # (and not a Bcc header)
        self.assertEqual(
            data["content"]["headers"],
            {"Cc": "Carbon Copy <cc1@example.com>, cc2@example.com"},
        )

    def test_custom_headers(self):
        self.set_mock_result(accepted=6)
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            cc=["cc1@example.com"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                "Message-ID": "mycustommsgid@example.com",
            },
        )
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["content"]["headers"],
            {
                # Reply-To moved to separate param (below)
                "X-MyHeader": "my value",
                "Message-ID": "mycustommsgid@example.com",
                "Cc": "cc1@example.com",  # Cc header added
            },
        )
        self.assertEqual(data["content"]["reply_to"], "another@example.com")

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(data["content"]["text"], text_content)
        self.assertEqual(data["content"]["html"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data["content"])

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()

        data = self.get_api_call_json()
        self.assertNotIn("text", data["content"])
        self.assertEqual(data["content"]["html"], html_content)

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
            headers={"X-Other": "Keep"},
        )
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["content"]["reply_to"], "reply@example.com, Other <reply2@example.com>"
        )
        # don't lose other headers:
        self.assertEqual(data["content"]["headers"], {"X-Other": "Keep"})

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf params"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["content"]["attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["type"], "text/plain")
        self.assertEqual(attachments[0]["name"], "test.txt")
        self.assertEqual(
            decode_att(attachments[0]["data"]).decode("ascii"), text_content
        )
        self.assertEqual(attachments[1]["type"], "image/png")  # inferred from filename
        self.assertEqual(attachments[1]["name"], "test.png")
        self.assertEqual(decode_att(attachments[1]["data"]), png_content)
        self.assertEqual(attachments[2]["type"], "application/pdf")
        self.assertEqual(attachments[2]["name"], "")  # none
        self.assertEqual(decode_att(attachments[2]["data"]), pdf_content)
        # Make sure the image attachment is not treated as embedded:
        self.assertNotIn("inline_images", data["content"])

    def test_unicode_attachment_correctly_decoded(self):
        # Slight modification from the Django unicode docs:
        # http://django.readthedocs.org/en/latest/ref/unicode.html#email
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        attachments = data["content"]["attachments"]
        self.assertEqual(len(attachments), 1)

    def test_attachment_charset(self):
        # SparkPost allows charset param in attachment type
        self.message.attach(MIMEText("Une pièce jointe", "plain", "iso8859-1"))
        self.message.send()
        data = self.get_api_call_json()
        attachment = data["content"]["attachments"][0]
        self.assertEqual(attachment["type"], 'text/plain; charset="iso8859-1"')
        self.assertEqual(
            decode_att(attachment["data"]), "Une pièce jointe".encode("iso8859-1")
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["content"]["html"], html_content)

        self.assertEqual(len(data["content"]["inline_images"]), 1)
        self.assertEqual(data["content"]["inline_images"][0]["type"], "image/png")
        self.assertEqual(data["content"]["inline_images"][0]["name"], cid)
        self.assertEqual(
            decode_att(data["content"]["inline_images"][0]["data"]), image_data
        )
        # Make sure neither the html nor the inline image is treated as an attachment:
        self.assertNotIn("attachments", data["content"])

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["content"]["attachments"]
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0]["type"], "image/png")
        self.assertEqual(attachments[0]["name"], image_filename)
        self.assertEqual(decode_att(attachments[0]["data"]), image_data)
        self.assertEqual(attachments[1]["type"], "image/png")
        self.assertEqual(attachments[1]["name"], "")  # unknown -- not attached as file
        self.assertEqual(decode_att(attachments[1]["data"]), image_data)
        # Make sure the image attachments are not treated as embedded:
        self.assertNotIn("inline_images", data["content"])

    def test_multiple_html_alternatives(self):
        # Multiple text/html alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_amp_html_alternative(self):
        # SparkPost *does* support text/x-amp-html alongside text/html
        self.message.attach_alternative("<p>HTML</p>", "text/html")
        self.message.attach_alternative("<p>And AMP HTML</p>", "text/x-amp-html")
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["content"]["html"], "<p>HTML</p>")
        self.assertEqual(data["content"]["amp_html"], "<p>And AMP HTML</p>")

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty headers"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("headers", data["content"])  # No Cc, Bcc or Reply-To header
        self.assertNotIn("reply_to", data["content"])

    def test_empty_to(self):
        # Test empty `to`--but send requires at least one recipient somewhere (like cc)
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["recipients"],
            [
                {
                    "address": {
                        "email": "cc@example.com",
                        # This results in a message with an empty To header, as desired:
                        "header_to": "",
                    },
                }
            ],
        )

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "SparkPost API response 400"):
            self.message.send()

    def test_api_failure_fail_silently(self):
        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        failure_response = b"""{
            "errors": [{
                "message": "Helpful explanation from your ESP"
            }]
        }"""
        self.set_mock_response(status_code=400, raw=failure_response)
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from your ESP"
        ):
            self.message.send()


@tag("sparkpost")
class SparkPostBackendAnymailFeatureTests(SparkPostBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "bounce-handler@bounces.example.com"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["return_path"], "bounce-handler@bounces.example.com")

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": "spark, post"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["metadata"], {"user_id": "12345", "items": "spark, post"})

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        # SparkPost expects ISO-8601 YYYY-MM-DDTHH:MM:SS+-HH:MM
        with override_current_timezone(utc_plus_6):
            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2016-03-04T05:06:07-08:00")

            # Explicit UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=timezone.utc)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2016-03-04T05:06:07+00:00")

            # Timezone-naive datetime assumed to be Django current_timezone
            # (also checks stripping microseconds)
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 567)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2022-10-11T12:13:14+06:00")

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2022-10-22T00:00:00+06:00")

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2022-05-06T07:08:09+00:00")

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "2022-10-13T18:02:00-11:30"
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["options"]["start_time"], "2022-10-13T18:02:00-11:30")

    def test_tags(self):
        self.message.tags = ["receipt"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["campaign_id"], "receipt")

        self.message.tags = ["receipt", "repeat-user"]
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple tags"):
            self.message.send()

    def test_tracking(self):
        # Test one way...
        self.message.track_opens = True
        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["options"]["open_tracking"], True)
        self.assertEqual(data["options"]["click_tracking"], False)

        # ...and the opposite way
        self.message.track_opens = False
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["options"]["open_tracking"], False)
        self.assertEqual(data["options"]["click_tracking"], True)

    def test_template_id(self):
        message = mail.EmailMultiAlternatives(to=["to@example.com"])
        message.from_email = None
        message.template_id = "welcome_template"
        message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["content"]["template_id"], "welcome_template")
        # SparkPost disallows all content (even empty strings) with stored template:
        self.assertNotIn("subject", data["content"])
        self.assertNotIn("text", data["content"])
        self.assertNotIn("html", data["content"])
        self.assertNotIn("from", data["content"])

    def test_template_id_ignores_default_from_email(self):
        # No from_email, or from_email=None in constructor,
        # uses settings.DEFAULT_FROM_EMAIL. We strip that with a template_id:
        message = AnymailMessage(to=["to@example.com"], template_id="welcome_template")
        self.assertIsNotNone(message.from_email)  # because it's DEFAULT_FROM_EMAIL
        message.send()
        data = self.get_api_call_json()
        self.assertNotIn("from", data["content"])

    def test_unsupported_content_with_template_id(self):
        # Make sure we raise an error for options that SparkPost
        # silently ignores when sending with a template_id
        message = AnymailMessage(
            to=["to@example.com"],
            from_email="non-default-from@example.com",
            reply_to=["reply@example.com"],
            headers={"X-Custom": "header"},
            template_id="welcome_template",
        )
        message.attach(filename="test.txt", content="attachment", mimetype="text/plain")
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature,
            "template_id with attachments, extra headers and/or cc recipients,"
            " from_email, reply_to",
        ):
            message.send()

    def test_merge_data(self):
        self.set_mock_result(accepted=4)  # two 'to' plus one 'cc' for each 'to'
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.cc = ["cc@example.com"]
        self.message.body = "Hi {{address.name}}. Welcome to {{group}} at {{site}}."
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
            "nobody@example.com": {"name": "Not a recipient for this message"},
        }
        self.message.merge_global_data = {"group": "Users", "site": "ExampleCo"}

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            {"group": "Users", "site": "ExampleCo"}, data["substitution_data"]
        )
        self.assertEqual(
            [
                {
                    "address": {
                        "email": "alice@example.com",
                        "header_to": "alice@example.com",
                    },
                    "substitution_data": {"name": "Alice", "group": "Developers"},
                },
                {
                    "address": {
                        "email": "bob@example.com",
                        "header_to": "Bob <bob@example.com>",
                    },
                    "substitution_data": {"name": "Bob"},
                },
                {  # duplicated for cc recipients...
                    "address": {
                        "email": "cc@example.com",
                        "header_to": "alice@example.com",
                    },
                    "substitution_data": {"name": "Alice", "group": "Developers"},
                },
                {
                    "address": {
                        "email": "cc@example.com",
                        "header_to": "Bob <bob@example.com>",
                    },
                    "substitution_data": {"name": "Bob"},
                },
            ],
            data["recipients"],
        )

    def test_merge_metadata(self):
        self.set_mock_result(accepted=2)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }
        self.message.metadata = {"notification_batch": "zx912"}

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            [
                {
                    "address": {
                        "email": "alice@example.com",
                        "header_to": "alice@example.com",
                    },
                    "metadata": {"order_id": 123},
                },
                {
                    "address": {
                        "email": "bob@example.com",
                        "header_to": "Bob <bob@example.com>",
                    },
                    "metadata": {"order_id": 678, "tier": "premium"},
                },
            ],
            data["recipients"],
        )
        self.assertEqual(data["metadata"], {"notification_batch": "zx912"})

    def test_merge_headers(self):
        self.set_mock_result(accepted=2)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "X-Custom-1": "custom 1",
            "X-Custom-2": "custom 2 (default)",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "X-Custom-2": "custom 2 alice",
                "X-Custom-3": "custom 3 alice",
            },
            "bob@example.com": {"X-Custom-2": "custom 2 bob"},
        }

        self.message.send()
        data = self.get_api_call_json()
        recipients = data["recipients"]
        self.assertEqual(len(recipients), 2)
        self.assertEqual(recipients[0]["address"]["email"], "alice@example.com")
        self.assertEqual(
            recipients[0]["substitution_data"],
            {
                "Header__X_Custom_2": "custom 2 alice",
                "Header__X_Custom_3": "custom 3 alice",
            },
        )
        self.assertEqual(recipients[1]["address"]["email"], "bob@example.com")
        self.assertEqual(
            recipients[1]["substitution_data"],
            {
                "Header__X_Custom_2": "custom 2 bob",
            },
        )
        # Indirect merge_headers through template substitutions:
        self.assertEqual(
            data["content"]["headers"],
            {
                "X-Custom-1": "custom 1",  # (not a merge_header, value unchanged)
                "X-Custom-2": "{{Header__X_Custom_2}}",
                "X-Custom-3": "{{Header__X_Custom_3}}",
            },
        )
        # Defaults for merge_headers in global substitution_data:
        self.assertEqual(
            data["substitution_data"],
            {
                "Header__X_Custom_2": "custom 2 (default)",
                # No default specified for X-Custom-3; SparkPost will use empty string
            },
        )

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("campaign_id", data)
        self.assertNotIn("metadata", data)
        # covers start_time, click_tracking, open_tracking:
        self.assertNotIn("options", data)
        self.assertNotIn("substitution_data", data)
        self.assertNotIn("template_id", data["content"])

    def test_esp_extra(self):
        self.message.esp_extra = {
            "description": "The description",
            "options": {
                "transactional": True,
            },
            "content": {
                "use_draft_template": True,
                "ab_test_id": "highlight_support_links",
            },
        }
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["description"], "The description")
        self.assertEqual(
            data["options"],
            {
                "transactional": True,
                "click_tracking": True,  # deep merge
            },
        )
        self.assertDictMatches(
            {
                "use_draft_template": True,
                "ab_test_id": "highlight_support_links",
                "text": "Text Body",  # deep merge
                "subject": "Subject",  # deep merge
            },
            data["content"],
        )

    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = self.set_mock_result(accepted=1, rejected=0, id="9876543210")
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, "9876543210")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, "9876543210"
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    @override_settings(
        ANYMAIL_IGNORE_RECIPIENT_STATUS=True  # exception is tested later
    )
    def test_send_all_rejected(self):
        """The anymail_status should be 'rejected' when all recipients rejected"""
        self.set_mock_result(accepted=0, rejected=2)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "to2@example.com"],
        )
        msg.send()
        self.assertEqual(msg.anymail_status.status, {"rejected"})
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "rejected"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to2@example.com"].status, "rejected"
        )

    def test_send_some_rejected(self):
        """
        The anymail_status should be 'unknown'
        when some recipients accepted and some rejected
        """
        self.set_mock_result(accepted=1, rejected=1)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "to2@example.com"],
        )
        msg.send()
        self.assertEqual(msg.anymail_status.status, {"unknown"})
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "unknown"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to2@example.com"].status, "unknown"
        )

    def test_send_unexpected_count(self):
        """The anymail_status should be 'unknown' when the total result count
        doesn't match the number of recipients"""
        self.set_mock_result(accepted=3, rejected=0)  # but only 2 in the to-list
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "to2@example.com"],
        )
        msg.send()
        self.assertEqual(msg.anymail_status.status, {"unknown"})
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "unknown"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to2@example.com"].status, "unknown"
        )

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=400)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but result is unexpected format,
        should raise an API exception
        """
        response_content = b"""{"wrong": "format"}"""
        self.set_mock_response(raw=response_content)
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(
            self.message.anymail_status.esp_response.content, response_content
        )

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.tags = [Decimal("19.99")]  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to SparkPost", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("sparkpost")
class SparkPostBackendRecipientsRefusedTests(SparkPostBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    def test_recipients_refused(self):
        self.set_mock_result(accepted=0, rejected=2)
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
        )
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()

    def test_fail_silently(self):
        self.set_mock_result(accepted=0, rejected=2)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_mixed_response(self):
        """If *any* recipients are valid or queued, no exception is raised"""
        self.set_mock_result(accepted=2, rejected=2)
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            [
                "invalid@localhost",
                "valid@example.com",
                "reject@example.com",
                "also.valid@example.com",
            ],
        )
        sent = msg.send()
        # one message sent, successfully, to 2 of 4 recipients:
        self.assertEqual(sent, 1)
        status = msg.anymail_status
        # We don't know which recipients were rejected
        self.assertEqual(status.recipients["invalid@localhost"].status, "unknown")
        self.assertEqual(status.recipients["valid@example.com"].status, "unknown")
        self.assertEqual(status.recipients["reject@example.com"].status, "unknown")
        self.assertEqual(status.recipients["also.valid@example.com"].status, "unknown")

    @override_settings(ANYMAIL_IGNORE_RECIPIENT_STATUS=True)
    def test_settings_override(self):
        """No exception with ignore setting"""
        self.set_mock_result(accepted=0, rejected=2)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
        )
        self.assertEqual(sent, 1)  # refused message is included in sent count


@tag("sparkpost")
class SparkPostBackendConfigurationTests(SparkPostBackendMockAPITestCase):
    """Test various SparkPost client options"""

    @override_settings(
        # clear SPARKPOST_API_KEY from SparkPostBackendMockAPITestCase:
        ANYMAIL={}
    )
    def test_missing_api_key(self):
        with self.assertRaises(AnymailConfigurationError) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        # Make sure the error mentions the different places to set the key
        self.assertRegex(errmsg, r"\bSPARKPOST_API_KEY\b")
        self.assertRegex(errmsg, r"\bANYMAIL_SPARKPOST_API_KEY\b")

    @override_settings(
        ANYMAIL={
            "SPARKPOST_API_URL": "https://api.eu.sparkpost.com/api/v1",
            "SPARKPOST_API_KEY": "test_api_key",
        }
    )
    def test_sparkpost_api_url(self):
        mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        self.assert_esp_called("https://api.eu.sparkpost.com/api/v1/transmissions/")

        # can also override on individual connection
        # (and even use non-versioned labs endpoint)
        connection = mail.get_connection(api_url="https://api.sparkpost.com/api/labs")
        mail.send_mail(
            "Subject",
            "Message",
            "from@example.com",
            ["to@example.com"],
            connection=connection,
        )
        self.assert_esp_called("https://api.sparkpost.com/api/labs/transmissions/")

    def test_subaccount(self):
        # A likely use case is supplying subaccount for a particular message.
        # (For all messages, just set SPARKPOST_SUBACCOUNT in ANYMAIL settings.)
        connection = mail.get_connection(subaccount=123)
        mail.send_mail(
            "Subject",
            "Message",
            "from@example.com",
            ["to@example.com"],
            connection=connection,
        )
        headers = self.get_api_call_headers()
        self.assertEqual(headers["X-MSYS-SUBACCOUNT"], 123)

        # Make sure we're not setting the header on non-subaccount sends
        mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        headers = self.get_api_call_headers()
        self.assertNotIn("X-MSYS-SUBACCOUNT", headers)



from datetime import date, datetime
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailRecipientsRefused,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("mandrill")
@override_settings(
    EMAIL_BACKEND="anymail.backends.mandrill.EmailBackend",
    ANYMAIL={"MANDRILL_API_KEY": "test_api_key"},
)
class MandrillBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""[{
        "email": "to@example.com",
        "status": "sent",
        "_id": "abc123",
        "reject_reason": null
    }]"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("mandrill")
class MandrillBackendStandardEmailTests(MandrillBackendMockAPITestCase):
    """Test backend support for Django mail wrappers"""

    def test_send_mail(self):
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/messages/send.json")
        data = self.get_api_call_json()
        self.assertEqual(data["key"], "test_api_key")
        self.assertEqual(data["message"]["subject"], "Subject here")
        self.assertEqual(data["message"]["text"], "Here is the message.")
        self.assertNotIn("from_name", data["message"])
        self.assertEqual(data["message"]["from_email"], "from@example.com")
        self.assertEqual(
            data["message"]["to"], [{"email": "to@example.com", "type": "to"}]
        )

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["from_name"], "From Name")
        self.assertEqual(data["message"]["from_email"], "from@example.com")
        self.assertEqual(
            data["message"]["to"],
            [
                {"email": "to1@example.com", "name": "Recipient #1", "type": "to"},
                {"email": "to2@example.com", "type": "to"},
                {"email": "cc1@example.com", "name": "Carbon Copy", "type": "cc"},
                {"email": "cc2@example.com", "type": "cc"},
                {"email": "bcc1@example.com", "name": "Blind Copy", "type": "bcc"},
                {"email": "bcc2@example.com", "type": "bcc"},
            ],
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                "Message-ID": "mycustommsgid@example.com",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["subject"], "Subject")
        self.assertEqual(data["message"]["text"], "Body goes here")
        self.assertEqual(data["message"]["from_email"], "from@example.com")
        self.assertEqual(
            data["message"]["headers"],
            {
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                "Message-ID": "mycustommsgid@example.com",
            },
        )
        # Verify recipients correctly identified as "to", "cc", or "bcc"
        self.assertEqual(
            data["message"]["to"],
            [
                {"email": "to1@example.com", "type": "to"},
                {"email": "to2@example.com", "name": "Also To", "type": "to"},
                {"email": "cc1@example.com", "type": "cc"},
                {"email": "cc2@example.com", "name": "Also CC", "type": "cc"},
                {"email": "bcc1@example.com", "type": "bcc"},
                {"email": "bcc2@example.com", "name": "Also BCC", "type": "bcc"},
            ],
        )
        # Don't use Mandrill's bcc_address "logging" feature for bcc's:
        self.assertNotIn("bcc_address", data["message"])

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["text"], text_content)
        self.assertEqual(data["message"]["html"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertFalse("attachments" in data["message"])

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertNotIn("text", data["message"])
        self.assertEqual(data["message"]["html"], html_content)

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
            headers={"X-Other": "Keep"},
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["message"]["headers"]["Reply-To"],
            "reply@example.com, Other <reply2@example.com>",
        )
        # don't lose other headers:
        self.assertEqual(data["message"]["headers"]["X-Other"], "Keep")

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["message"]["attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["type"], "text/plain")
        self.assertEqual(attachments[0]["name"], "test.txt")
        self.assertEqual(
            decode_att(attachments[0]["content"]).decode("ascii"), text_content
        )
        self.assertEqual(attachments[1]["type"], "image/png")  # inferred from filename
        self.assertEqual(attachments[1]["name"], "test.png")
        self.assertEqual(decode_att(attachments[1]["content"]), png_content)
        self.assertEqual(attachments[2]["type"], "application/pdf")
        self.assertEqual(attachments[2]["name"], "")  # none
        self.assertEqual(decode_att(attachments[2]["content"]), pdf_content)
        # Make sure the image attachment is not treated as embedded:
        self.assertFalse("images" in data["message"])

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        attachments = data["message"]["attachments"]
        self.assertEqual(len(attachments), 1)

    def test_embedded_images(self):
        image_data = sample_image_content()  # Read from a png file

        cid = attach_inline_image(self.message, image_data)
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["message"]["images"]), 1)
        self.assertEqual(data["message"]["images"][0]["type"], "image/png")
        self.assertEqual(data["message"]["images"][0]["name"], cid)
        self.assertEqual(
            decode_att(data["message"]["images"][0]["content"]), image_data
        )
        # Make sure neither the html nor the inline image is treated as an attachment:
        self.assertFalse("attachments" in data["message"])

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file:
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly:
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["message"]["attachments"]
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0]["type"], "image/png")
        self.assertEqual(attachments[0]["name"], image_filename)
        self.assertEqual(decode_att(attachments[0]["content"]), image_data)
        self.assertEqual(attachments[1]["type"], "image/png")
        self.assertEqual(attachments[1]["name"], "")  # unknown -- not attached as file
        self.assertEqual(decode_att(attachments[1]["content"]), image_data)
        # Make sure the image attachments are not treated as embedded:
        self.assertFalse("images" in data["message"])

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "Mandrill API response 400"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        self.set_mock_response(
            status_code=400,
            raw=b"""{
             "status": "error",
             "code": 12,
             "name": "Error_Name",
             "message": "Helpful explanation from Mandrill"
        }""",
        )
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from Mandrill"
        ):
            self.message.send()

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Invalid API key")
        with self.assertRaisesMessage(AnymailAPIError, "Invalid API key"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("mandrill")
class MandrillBackendAnymailFeatureTests(MandrillBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "anything@bounces.example.com"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["return_path_domain"], "bounces.example.com")

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["metadata"], {"user_id": "12345", "items": 6})

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 567)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], "2022-10-11 06:13:14"
            )  # 12:13 UTC+6 == 06:13 UTC

            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], "2016-03-04 13:06:07"
            )  # 05:06 UTC-8 == 13:06 UTC

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], "2022-10-21 18:00:00"
            )  # 00:00 UTC+6 == 18:00-1d UTC

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["send_at"], "2022-05-06 07:08:09")

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "2013-11-12 01:02:03"
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["send_at"], "2013-11-12 01:02:03")

    def test_tags(self):
        self.message.tags = ["receipt", "repeat-user"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["tags"], ["receipt", "repeat-user"])

    def test_tracking(self):
        # Test one way...
        self.message.track_opens = True
        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["track_opens"], True)
        self.assertEqual(data["message"]["track_clicks"], False)

        # ...and the opposite way
        self.message.track_opens = False
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["track_opens"], False)
        self.assertEqual(data["message"]["track_clicks"], True)

    def test_template_id(self):
        self.message.template_id = "welcome_template"
        self.message.send()
        data = self.get_api_call_json()
        # template requires different send API:
        self.assert_esp_called("/messages/send-template.json")
        self.assertEqual(data["template_name"], "welcome_template")
        # Mandrill requires template_content with send-template:
        self.assertEqual(data["template_content"], [])

    def test_merge_data(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        # Mandrill template_id is not required to use merge.
        # You can just supply template content as the message (e.g.):
        self.message.body = "Hi *|name|*. Welcome to *|group|* at *|site|*."
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave :group undefined
        }
        self.message.merge_global_data = {
            "group": "Users",
            "site": "ExampleCo",
        }
        self.message.send()
        # didn't specify template_id, so use normal send:
        self.assert_esp_called("/messages/send.json")
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["message"]["merge_vars"],
            [
                {
                    "rcpt": "alice@example.com",
                    "vars": [
                        {"name": "group", "content": "Developers"},
                        {"name": "name", "content": "Alice"},
                    ],
                },
                {
                    "rcpt": "bob@example.com",
                    "vars": [{"name": "name", "content": "Bob"}],
                },
            ],
        )
        self.assertCountEqual(
            data["message"]["global_merge_vars"],
            [
                {"name": "group", "content": "Users"},
                {"name": "site", "content": "ExampleCo"},
            ],
        )
        # merge_data implies batch:
        self.assertIs(data["message"]["preserve_recipients"], False)

    def test_merge_metadata(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},
        }
        self.message.metadata = {"notification_batch": "zx912"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["message"]["recipient_metadata"],
            [
                {
                    "rcpt": "alice@example.com",
                    "values": {"order_id": 123, "tier": "premium"},
                },
                {
                    "rcpt": "bob@example.com",
                    "values": {"order_id": 678},
                },
            ],
        )
        # merge_metadata implies batch:
        self.assertIs(data["message"]["preserve_recipients"], False)

    def test_missing_from(self):
        """Make sure a missing from_email omits from* from API call.

        (Allows use of from email/name from template)
        """
        # You must set from_email=None after constructing the EmailMessage
        # (or you will end up with Django's settings.DEFAULT_FROM_EMAIL instead)
        self.message.from_email = None
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("from_email", data["message"])
        self.assertNotIn("from_name", data["message"])

    def test_missing_subject(self):
        """Make sure a missing subject omits subject from API call.

        (Allows use of template subject)
        """
        self.message.subject = None
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("subject", data["message"])

    def test_esp_extra(self):
        self.message.esp_extra = {
            # Mandrill send param that goes at top level of API payload:
            "ip_pool": "Bulk Pool",
            "message": {
                # param that goes within message dict:
                "subaccount": "Marketing Dept."
            },
        }
        self.message.tags = ["test-tag"]  # make sure non-esp_extra params are merged
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["ip_pool"], "Bulk Pool")
        self.assertEqual(data["message"]["subaccount"], "Marketing Dept.")
        self.assertEqual(data["message"]["tags"], ["test-tag"])

    def test_esp_extra_recipient_metadata(self):
        """Anymail allows pythonic recipient_metadata dict"""
        self.message.esp_extra = {
            "message": {
                "recipient_metadata": {
                    # Anymail expands simple python dicts into the more-verbose
                    # rcpt/values lists the Mandrill API uses
                    "customer@example.com": {"cust_id": "67890", "order_id": "54321"},
                    "guest@example.com": {"cust_id": "94107", "order_id": "43215"},
                }
            }
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["message"]["recipient_metadata"],
            [
                {
                    "rcpt": "customer@example.com",
                    "values": {"cust_id": "67890", "order_id": "54321"},
                },
                {
                    "rcpt": "guest@example.com",
                    "values": {"cust_id": "94107", "order_id": "43215"},
                },
            ],
        )

        # You can also just supply it in Mandrill's native form
        self.message.esp_extra = {
            "message": {
                "recipient_metadata": [
                    {
                        "rcpt": "customer@example.com",
                        "values": {"cust_id": "80806", "order_id": "70701"},
                    },
                    {
                        "rcpt": "guest@example.com",
                        "values": {"cust_id": "21212", "order_id": "10305"},
                    },
                ]
            }
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["message"]["recipient_metadata"],
            [
                {
                    "rcpt": "customer@example.com",
                    "values": {"cust_id": "80806", "order_id": "70701"},
                },
                {
                    "rcpt": "guest@example.com",
                    "values": {"cust_id": "21212", "order_id": "10305"},
                },
            ],
        )

    def test_esp_extra_template_content(self):
        """Anymail allows pythonic template_content dict"""
        self.message.template_id = (
            "welcome_template"  # forces send-template API and default template_content
        )
        self.message.esp_extra = {
            "template_content": {
                # Anymail expands simple python dicts into the more-verbose name/content
                # structures the Mandrill API uses
                "HEADLINE": "<h1>Specials Just For *|FNAME|*</h1>",
                "OFFER_BLOCK": "<p><em>Half off</em> all fruit</p>",
            }
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["template_content"],
            [
                {"name": "HEADLINE", "content": "<h1>Specials Just For *|FNAME|*</h1>"},
                {
                    "name": "OFFER_BLOCK",
                    "content": "<p><em>Half off</em> all fruit</p>",
                },
            ],
        )

        # You can also just supply it in Mandrill's native form
        self.message.esp_extra = {
            "template_content": [
                {
                    "name": "HEADLINE",
                    "content": "<h1>Exciting offers for *|FNAME|*</h1>",
                },
                {"name": "OFFER_BLOCK", "content": "<p><em>25% off</em> all fruit</p>"},
            ]
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["template_content"],
            [
                {
                    "name": "HEADLINE",
                    "content": "<h1>Exciting offers for *|FNAME|*</h1>",
                },
                {"name": "OFFER_BLOCK", "content": "<p><em>25% off</em> all fruit</p>"},
            ],
        )

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        self.assert_esp_called("/messages/send.json")
        data = self.get_api_call_json()
        self.assertNotIn("global_merge_vars", data["message"])
        self.assertNotIn("merge_vars", data["message"])
        self.assertNotIn("metadata", data["message"])
        self.assertNotIn("send_at", data)
        self.assertNotIn("tags", data["message"])
        self.assertNotIn("template_content", data["message"])
        self.assertNotIn("template_name", data["message"])
        self.assertNotIn("track_clicks", data["message"])
        self.assertNotIn("track_opens", data["message"])

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = (
            b'[{"email": "to1@example.com", "status": "sent", "_id": "abc123"}]'
        )
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"sent"})
        self.assertEqual(msg.anymail_status.message_id, "abc123")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "sent"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, "abc123"
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.metadata = {"total": Decimal("19.99")}
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_data())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Mandrill", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")

    def test_no_extraneous_fields(self):
        """Don't send empty fields that have no effect on sending"""
        mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])
        data = self.get_api_call_json()
        # Simple send should contain exactly this, nothing more:
        self.assertEqual(
            data,
            {
                "key": "test_api_key",
                "message": {
                    "from_email": "from@example.com",
                    "to": [{"email": "to@example.com", "type": "to"}],
                    "subject": "Subject",
                    "text": "Body",
                },
            },
        )


@tag("mandrill")
class MandrillBackendRecipientsRefusedTests(MandrillBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    def test_recipients_refused(self):
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@test.mandrillapp.com"],
        )
        self.set_mock_response(
            raw=b"""[
            {"email": "invalid@localhost", "status": "invalid"},
            {"email": "reject@test.mandrillapp.com", "status": "rejected"}
        ]"""
        )
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()

    def test_fail_silently(self):
        self.set_mock_response(
            raw=b"""[
            {"email": "invalid@localhost", "status": "invalid"},
            {"email": "reject@test.mandrillapp.com", "status": "rejected"}
        ]"""
        )
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@test.mandrillapp.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_mixed_response(self):
        """If *any* recipients are valid or queued, no exception is raised"""
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            [
                "invalid@localhost",
                "valid@example.com",
                "reject@test.mandrillapp.com",
                "also.valid@example.com",
            ],
        )
        self.set_mock_response(
            raw=b"""[
            {"email": "invalid@localhost", "status": "invalid"},
            {"email": "valid@example.com", "status": "sent"},
            {"email": "reject@test.mandrillapp.com", "status": "rejected"},
            {"email": "also.valid@example.com", "status": "queued"}
        ]"""
        )
        sent = msg.send()
        # one message sent, successfully, to 2 of 4 recipients:
        self.assertEqual(sent, 1)
        status = msg.anymail_status
        self.assertEqual(status.recipients["invalid@localhost"].status, "invalid")
        self.assertEqual(status.recipients["valid@example.com"].status, "sent")
        self.assertEqual(
            status.recipients["reject@test.mandrillapp.com"].status, "rejected"
        )
        self.assertEqual(status.recipients["also.valid@example.com"].status, "queued")

    @override_settings(ANYMAIL_IGNORE_RECIPIENT_STATUS=True)
    def test_settings_override(self):
        """No exception with ignore setting"""
        self.set_mock_response(
            raw=b"""[
            {"email": "invalid@localhost", "status": "invalid"},
            {"email": "reject@test.mandrillapp.com", "status": "rejected"}
        ]"""
        )
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@test.mandrillapp.com"],
        )
        self.assertEqual(sent, 1)  # refused message is included in sent count


@tag("mandrill")
class MandrillBackendSessionSharingTestCase(
    SessionSharingTestCases, MandrillBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("mandrill")
@override_settings(EMAIL_BACKEND="anymail.backends.mandrill.EmailBackend")
class MandrillBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test backend without required settings"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bMANDRILL_API_KEY\b")
        self.assertRegex(errmsg, r"\bANYMAIL_MANDRILL_API_KEY\b")



import mimetypes

from ..exceptions import AnymailRequestsAPIError
from ..message import AnymailRecipientStatus
from ..utils import (
    BASIC_NUMERIC_TYPES,
    CaseInsensitiveCasePreservingDict,
    get_anymail_setting,
)
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Resend (resend.com) API Email Backend
    """

    esp_name = "Resend"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://api.resend.com/",
        )
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return ResendPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)
        try:
            message_id = parsed_response["id"]
            message_ids = None
        except (KeyError, TypeError):
            # Batch send?
            try:
                message_id = None
                message_ids = [item["id"] for item in parsed_response["data"]]
            except (KeyError, TypeError) as err:
                raise AnymailRequestsAPIError(
                    "Invalid Resend API response format",
                    email_message=message,
                    payload=payload,
                    response=response,
                    backend=self,
                ) from err

        recipient_status = CaseInsensitiveCasePreservingDict(
            {
                recip.addr_spec: AnymailRecipientStatus(
                    message_id=message_id, status="queued"
                )
                for recip in payload.recipients
            }
        )
        if message_ids:
            # batch send: ids are in same order as to_recipients
            for recip, message_id in zip(payload.to_recipients, message_ids):
                recipient_status[recip.addr_spec] = AnymailRecipientStatus(
                    message_id=message_id, status="queued"
                )
        return dict(recipient_status)


class ResendPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        self.recipients = []  # for parse_recipient_status
        self.to_recipients = []  # for parse_recipient_status
        self.metadata = {}
        self.merge_metadata = {}
        self.merge_headers = {}
        headers = kwargs.pop("headers", {})
        headers["Authorization"] = "Bearer %s" % backend.api_key
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        super().__init__(message, defaults, backend, headers=headers, *args, **kwargs)

    def get_api_endpoint(self):
        if self.is_batch():
            return "emails/batch"
        return "emails"

    def serialize_data(self):
        payload = self.data
        if self.is_batch():
            # Burst payload across to addresses
            to_emails = self.data.pop("to", [])
            payload = []
            for to_email, to in zip(to_emails, self.to_recipients):
                data = self.data.copy()
                data["to"] = [to_email]
                if to.addr_spec in self.merge_metadata:
                    # Merge global metadata with any per-recipient metadata.
                    recipient_metadata = self.metadata.copy()
                    recipient_metadata.update(self.merge_metadata[to.addr_spec])
                    if "headers" in data:
                        data["headers"] = data["headers"].copy()
                    else:
                        data["headers"] = {}
                    data["headers"]["X-Metadata"] = self.serialize_json(
                        recipient_metadata
                    )
                if to.addr_spec in self.merge_headers:
                    if "headers" in data:
                        # Merge global headers (or X-Metadata from above)
                        headers = CaseInsensitiveCasePreservingDict(data["headers"])
                        headers.update(self.merge_headers[to.addr_spec])
                    else:
                        headers = self.merge_headers[to.addr_spec]
                    data["headers"] = headers
                payload.append(data)

        return self.serialize_json(payload)

    #
    # Payload construction
    #

    def init_payload(self):
        self.data = {}  # becomes json

    def set_from_email(self, email):
        self.data["from"] = email.address

    def set_recipients(self, recipient_type, emails):
        assert recipient_type in ["to", "cc", "bcc"]
        if emails:
            field = recipient_type
            self.data[field] = [email.address for email in emails]
            self.recipients += emails
            if recipient_type == "to":
                self.to_recipients = emails

    def set_subject(self, subject):
        self.data["subject"] = subject

    def set_reply_to(self, emails):
        if emails:
            self.data["reply_to"] = [email.address for email in emails]

    def set_extra_headers(self, headers):
        # Resend requires header values to be strings (not integers) as of 2023-10-20.
        # Stringify ints and floats; anything else is the caller's responsibility.
        self.data.setdefault("headers", {}).update(
            {
                k: str(v) if isinstance(v, BASIC_NUMERIC_TYPES) else v
                for k, v in headers.items()
            }
        )

    def set_text_body(self, body):
        self.data["text"] = body

    def set_html_body(self, body):
        if "html" in self.data:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["html"] = body

    @staticmethod
    def make_attachment(attachment):
        """Returns Resend attachment dict for attachment"""
        filename = attachment.name or ""
        if not filename:
            # Provide default name with reasonable extension.
            # (Resend guesses content type from the filename extension;
            # there doesn't seem to be any other way to specify it.)
            ext = mimetypes.guess_extension(attachment.content_type)
            if ext is not None:
                filename = f"attachment{ext}"
        att = {"content": attachment.b64content, "filename": filename}
        # attachment.inline / attachment.cid not supported
        return att

    def set_attachments(self, attachments):
        if attachments:
            if any(att.content_id for att in attachments):
                self.unsupported_feature("inline content-id")
            self.data["attachments"] = [
                self.make_attachment(attachment) for attachment in attachments
            ]

    def set_metadata(self, metadata):
        # Send metadata as json in a custom X-Metadata header.
        # (Resend's own "tags" are severely limited in character set)
        self.data.setdefault("headers", {})["X-Metadata"] = self.serialize_json(
            metadata
        )
        self.metadata = metadata  # may be needed for batch send in serialize_data

    def set_send_at(self, send_at):
        try:
            # Resend can't handle microseconds; truncate to milliseconds if necessary.
            send_at = send_at.isoformat(
                timespec="milliseconds" if send_at.microsecond else "seconds"
            )
        except AttributeError:
            # User is responsible for formatting their own string
            pass
        self.data["scheduled_at"] = send_at

    def set_tags(self, tags):
        # Send tags using a custom X-Tags header.
        # (Resend's own "tags" are severely limited in character set)
        self.data.setdefault("headers", {})["X-Tags"] = self.serialize_json(tags)

    # Resend doesn't support changing click/open tracking per message
    # def set_track_clicks(self, track_clicks):
    # def set_track_opens(self, track_opens):

    # Resend doesn't support server-rendered templates.
    # (Their template feature is rendered client-side,
    # using React in node.js.)
    # def set_template_id(self, template_id):
    # def set_merge_global_data(self, merge_global_data):

    def set_merge_data(self, merge_data):
        # Empty merge_data is a request to use batch send;
        # any other merge_data is unsupported.
        if any(recipient_data for recipient_data in merge_data.values()):
            self.unsupported_feature("merge_data")

    def set_merge_metadata(self, merge_metadata):
        self.merge_metadata = merge_metadata  # late bound in serialize_data

    def set_merge_headers(self, merge_headers):
        self.merge_headers = merge_headers  # late bound in serialize_data

    def set_esp_extra(self, extra):
        self.data.update(extra)



import os
import unittest
from datetime import datetime, timedelta
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin

ANYMAIL_TEST_BREVO_API_KEY = os.getenv("ANYMAIL_TEST_BREVO_API_KEY")
ANYMAIL_TEST_BREVO_DOMAIN = os.getenv("ANYMAIL_TEST_BREVO_DOMAIN")


@tag("brevo", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_BREVO_API_KEY and ANYMAIL_TEST_BREVO_DOMAIN,
    "Set ANYMAIL_TEST_BREVO_API_KEY and ANYMAIL_TEST_BREVO_DOMAIN "
    "environment variables to run Brevo integration tests",
)
@override_settings(
    ANYMAIL_BREVO_API_KEY=ANYMAIL_TEST_BREVO_API_KEY,
    ANYMAIL_BREVO_SEND_DEFAULTS=dict(),
    EMAIL_BACKEND="anymail.backends.brevo.EmailBackend",
)
class BrevoBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Brevo v3 API integration tests

    Brevo doesn't have sandbox so these tests run
    against the **live** Brevo API, using the
    environment variable `ANYMAIL_TEST_BREVO_API_KEY` as the API key,
    and `ANYMAIL_TEST_BREVO_DOMAIN` to construct sender addresses.
    If those variables are not set, these tests won't run.

    https://developers.brevo.com/docs/faq#how-can-i-test-the-api

    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_BREVO_DOMAIN
        self.message = AnymailMessage(
            "Anymail Brevo integration test",
            "Text content",
            self.from_email,
            ["test+to1@anymail.dev"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Brevo send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["test+to1@anymail.dev"].status
        message_id = anymail_status.recipients["test+to1@anymail.dev"].message_id

        self.assertEqual(sent_status, "queued")  # Brevo always queues
        # Message-ID can be ...@smtp-relay.mail.fr or .sendinblue.com:
        self.assertRegex(message_id, r"\<.+@.+\>")
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now() + timedelta(minutes=2)
        message = AnymailMessage(
            subject="Anymail Brevo all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            to=["test+to1@anymail.dev", '"Recipient 2, OK?" <test+to2@anymail.dev>'],
            cc=["test+cc1@anymail.dev", "Copy 2 <test+cc2@anymail.dev>"],
            bcc=["test+bcc1@anymail.dev", "Blind Copy 2 <test+bcc2@anymail.dev>"],
            # Brevo API v3 only supports single reply-to
            reply_to=['"Reply, with comma" <reply@example.com>'],
            headers={"X-Anymail-Test": "value", "X-Anymail-Count": 3},
            metadata={"meta1": "simple string", "meta2": 2},
            send_at=send_at,
            tags=["tag 1", "tag 2"],
        )
        # Brevo requires an HTML body:
        message.attach_alternative("<p>HTML content</p>", "text/html")

        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")

        message.send()
        # Brevo always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        self.assertRegex(message.anymail_status.message_id, r"\<.+@.+\>")

    def test_template(self):
        message = AnymailMessage(
            # There is a *new-style* template with this id in the Anymail test account:
            template_id=5,
            # Override template sender:
            from_email=formataddr(("Sender", self.from_email)),
            to=["Recipient 1 <test+to1@anymail.dev>", "test+to2@anymail.dev"],
            reply_to=["Do not reply <reply@example.dev>"],
            tags=["using-template"],
            # The Anymail test template includes `{{ params.SHIP_DATE }}`
            # and `{{ params.ORDER_ID }}` substitutions
            merge_data={
                "test+to1@anymail.dev": {"ORDER_ID": "12345"},
                "test+to2@anymail.dev": {"ORDER_ID": "23456"},
            },
            merge_global_data={"SHIP_DATE": "yesterday"},
            metadata={"customer-id": "unknown", "meta2": 2},
            merge_metadata={
                "test+to1@anymail.dev": {"customer-id": "ZXK9123"},
                "test+to2@anymail.dev": {"customer-id": "ZZT4192"},
            },
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "test+to1@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "test+to2@anymail.dev": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )

        message.attach("attachment1.txt", "Here is some\ntext", "text/plain")

        message.send()
        # Brevo always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["test+to1@anymail.dev"].status, "queued")
        self.assertEqual(recipient_status["test+to2@anymail.dev"].status, "queued")
        self.assertRegex(
            recipient_status["test+to1@anymail.dev"].message_id, r"\<.+@.+\>"
        )
        self.assertRegex(
            recipient_status["test+to2@anymail.dev"].message_id, r"\<.+@.+\>"
        )
        # Each recipient gets their own message_id:
        self.assertNotEqual(
            recipient_status["test+to1@anymail.dev"].message_id,
            recipient_status["test+to2@anymail.dev"].message_id,
        )

    @override_settings(ANYMAIL_BREVO_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        # Make sure the exception message includes Brevo's response:
        self.assertIn("Key not found", str(err))



from base64 import b64encode

from django.test import override_settings

from tests.utils import ClientWithCsrfChecks

HAS_CRYPTOGRAPHY = True
try:
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import padding, rsa
except ImportError:
    HAS_CRYPTOGRAPHY = False


def make_key():
    """Generate RSA public key with short key size, for testing only"""
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
    )
    return private_key


def derive_public_webhook_key(private_key):
    """Derive public"""
    public_key = private_key.public_key()
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    public_bytes = b"\n".join(public_bytes.splitlines()[1:-1])
    return public_bytes.decode("utf-8")


def sign(private_key, message):
    """Sign message with private key"""
    signature = private_key.sign(message, padding.PKCS1v15(), hashes.SHA1())
    return signature


class _ClientWithPostalSignature(ClientWithCsrfChecks):
    private_key = None

    def set_private_key(self, private_key):
        self.private_key = private_key

    def post(self, *args, **kwargs):
        signature = b64encode(sign(self.private_key, kwargs["data"].encode("utf-8")))
        kwargs.setdefault("HTTP_X_POSTAL_SIGNATURE", signature)

        webhook_key = derive_public_webhook_key(self.private_key)
        with override_settings(ANYMAIL={"POSTAL_WEBHOOK_KEY": webhook_key}):
            return super().post(*args, **kwargs)


ClientWithPostalSignature = _ClientWithPostalSignature if HAS_CRYPTOGRAPHY else None



import binascii
import json
from base64 import b64decode
from datetime import datetime, timezone

from ..exceptions import (
    AnymailConfigurationError,
    AnymailImproperlyInstalled,
    AnymailInvalidAddress,
    AnymailWebhookValidationFailure,
    _LazyError,
)
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import get_anymail_setting, parse_single_address
from .base import AnymailBaseWebhookView

try:
    from cryptography.exceptions import InvalidSignature
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import padding
except ImportError:
    # This module gets imported by anymail.urls, so don't complain about cryptography
    # missing unless one of the Postal webhook views is actually used and needs it
    error = _LazyError(
        AnymailImproperlyInstalled(
            missing_package="cryptography", install_extra="postal"
        )
    )
    serialization = error
    hashes = error
    default_backend = error
    padding = error
    InvalidSignature = object


class PostalBaseWebhookView(AnymailBaseWebhookView):
    """Base view class for Postal webhooks"""

    esp_name = "Postal"

    warn_if_no_basic_auth = False

    # These can be set from kwargs in View.as_view, or pulled from settings in init:
    webhook_key = None

    def __init__(self, **kwargs):
        self.webhook_key = get_anymail_setting(
            "webhook_key", esp_name=self.esp_name, kwargs=kwargs, allow_bare=True
        )

        super().__init__(**kwargs)

    def validate_request(self, request):
        try:
            signature = request.META["HTTP_X_POSTAL_SIGNATURE"]
        except KeyError:
            raise AnymailWebhookValidationFailure(
                "X-Postal-Signature header missing from webhook"
            )

        public_key = serialization.load_pem_public_key(
            (
                "-----BEGIN PUBLIC KEY-----\n"
                + self.webhook_key
                + "\n-----END PUBLIC KEY-----"
            ).encode(),
            backend=default_backend(),
        )

        try:
            public_key.verify(
                b64decode(signature), request.body, padding.PKCS1v15(), hashes.SHA1()
            )
        except (InvalidSignature, binascii.Error):
            raise AnymailWebhookValidationFailure(
                "Postal webhook called with incorrect signature"
            )


class PostalTrackingWebhookView(PostalBaseWebhookView):
    """Handler for Postal message, engagement, and generation event webhooks"""

    signal = tracking

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))

        if "rcpt_to" in esp_event:
            raise AnymailConfigurationError(
                "You seem to have set Postal's *inbound* webhook "
                "to Anymail's Postal *tracking* webhook URL."
            )

        raw_timestamp = esp_event.get("timestamp")
        timestamp = (
            datetime.fromtimestamp(int(raw_timestamp), tz=timezone.utc)
            if raw_timestamp
            else None
        )

        payload = esp_event.get("payload", {})

        status_types = {
            "Sent": EventType.DELIVERED,
            "SoftFail": EventType.DEFERRED,
            "HardFail": EventType.FAILED,
            "Held": EventType.QUEUED,
        }

        if "status" in payload:
            event_type = status_types.get(payload["status"], EventType.UNKNOWN)
        elif "bounce" in payload:
            event_type = EventType.BOUNCED
        elif "url" in payload:
            event_type = EventType.CLICKED
        else:
            event_type = EventType.UNKNOWN

        description = payload.get("details")
        mta_response = payload.get("output")

        # extract message-related fields
        message = payload.get("message") or payload.get("original_message", {})
        message_id = message.get("id")
        tag = message.get("tag")
        recipient = None
        message_to = message.get("to")
        if message_to is not None:
            try:
                recipient = parse_single_address(message_to).addr_spec
            except AnymailInvalidAddress:
                pass

        if message.get("direction") == "incoming":
            # Let's ignore tracking events about an inbound emails.
            # This happens when an inbound email could not be forwarded.
            # The email didn't originate from Anymail, so the user can't do much about
            # it. It is part of normal Postal operation, not a configuration error.
            return []

        # only for MessageLinkClicked
        click_url = payload.get("url")
        user_agent = payload.get("user_agent")

        event = AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            event_id=esp_event.get("uuid"),
            esp_event=esp_event,
            click_url=click_url,
            description=description,
            message_id=message_id,
            metadata=None,
            mta_response=mta_response,
            recipient=recipient,
            reject_reason=(
                RejectReason.BOUNCED if event_type == EventType.BOUNCED else None
            ),
            tags=[tag],
            user_agent=user_agent,
        )

        return [event]


class PostalInboundWebhookView(PostalBaseWebhookView):
    """Handler for Postal inbound relay webhook"""

    signal = inbound

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))

        if "status" in esp_event:
            raise AnymailConfigurationError(
                "You seem to have set Postal's *tracking* webhook "
                "to Anymail's Postal *inbound* webhook URL."
            )

        raw_mime = esp_event["message"]
        if esp_event.get("base64") is True:
            raw_mime = b64decode(esp_event["message"]).decode("utf-8")
        message = AnymailInboundMessage.parse_raw_mime(raw_mime)

        message.envelope_sender = esp_event.get("mail_from", None)
        message.envelope_recipient = esp_event.get("rcpt_to", None)

        event = AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=None,
            event_id=esp_event.get("id"),
            esp_event=esp_event,
            message=message,
        )

        return [event]



import warnings

from ..exceptions import AnymailDeprecationWarning
from .brevo import EmailBackend as BrevoEmailBackend


class EmailBackend(BrevoEmailBackend):
    """
    Deprecated compatibility backend for old Brevo name "SendinBlue".
    """

    esp_name = "SendinBlue"

    def __init__(self, **kwargs):
        warnings.warn(
            "`anymail.backends.sendinblue.EmailBackend` has been renamed"
            " `anymail.backends.brevo.EmailBackend`.",
            AnymailDeprecationWarning,
        )
        super().__init__(**kwargs)



# These are the default "django-admin startproject" settings.py files
# for each supported version of Django, with:
#
#   * "anymail" appended to INSTALLED_APPS
#   * ROOT_URLCONF = 'tests.test_settings.urls'
#
# Keeping a file for each Django version is simpler than trying
# to maintain warning-free compatibility with all Django versions
# in a single settings list.
#
# It also helps ensure compatibility with default apps, middleware, etc.
#
# You can typically find the default settings.py template in
# SITE_PACKAGES/django/conf/project_template/project_name/settings.py-tpl



from ..exceptions import AnymailRequestsAPIError
from ..message import AnymailRecipientStatus
from ..utils import get_anymail_setting
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Postal v1 API Email Backend
    """

    esp_name = "Postal"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name

        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )

        # Required, as there is no hosted instance of Postal
        api_url = get_anymail_setting("api_url", esp_name=esp_name, kwargs=kwargs)
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return PostalPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)

        if parsed_response["status"] != "success":
            raise AnymailRequestsAPIError(
                email_message=message, payload=payload, response=response, backend=self
            )

        # If we get here, the send call was successful.
        messages = parsed_response["data"]["messages"]

        return {
            email: AnymailRecipientStatus(message_id=details["id"], status="queued")
            for email, details in messages.items()
        }


class PostalPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        http_headers = kwargs.pop("headers", {})
        http_headers["X-Server-API-Key"] = backend.api_key
        http_headers["Content-Type"] = "application/json"
        http_headers["Accept"] = "application/json"
        super().__init__(
            message, defaults, backend, headers=http_headers, *args, **kwargs
        )

    def get_api_endpoint(self):
        return "api/v1/send/message"

    def init_payload(self):
        self.data = {}

    def serialize_data(self):
        return self.serialize_json(self.data)

    def set_from_email(self, email):
        self.data["from"] = str(email)

    def set_subject(self, subject):
        self.data["subject"] = subject

    def set_to(self, emails):
        self.data["to"] = [str(email) for email in emails]

    def set_cc(self, emails):
        self.data["cc"] = [str(email) for email in emails]

    def set_bcc(self, emails):
        self.data["bcc"] = [str(email) for email in emails]

    def set_reply_to(self, emails):
        if len(emails) > 1:
            self.unsupported_feature("multiple reply_to addresses")
        if len(emails) > 0:
            self.data["reply_to"] = str(emails[0])

    def set_extra_headers(self, headers):
        self.data["headers"] = headers

    def set_text_body(self, body):
        self.data["plain_body"] = body

    def set_html_body(self, body):
        if "html_body" in self.data:
            self.unsupported_feature("multiple html parts")
        self.data["html_body"] = body

    def make_attachment(self, attachment):
        """Returns Postal attachment dict for attachment"""
        att = {
            "name": attachment.name or "",
            "data": attachment.b64content,
            "content_type": attachment.mimetype,
        }
        if attachment.inline:
            # see https://github.com/postalhq/postal/issues/731
            # but it might be possible with the send/raw endpoint
            self.unsupported_feature("inline attachments")
        return att

    def set_attachments(self, attachments):
        if attachments:
            self.data["attachments"] = [
                self.make_attachment(attachment) for attachment in attachments
            ]

    def set_envelope_sender(self, email):
        self.data["sender"] = str(email)

    def set_tags(self, tags):
        if len(tags) > 1:
            self.unsupported_feature("multiple tags")
        if len(tags) > 0:
            self.data["tag"] = tags[0]

    def set_esp_extra(self, extra):
        self.data.update(extra)



import json
from datetime import datetime

from ..exceptions import (
    AnymailImproperlyInstalled,
    AnymailInvalidAddress,
    AnymailWebhookValidationFailure,
    _LazyError,
)
from ..signals import AnymailTrackingEvent, EventType, RejectReason, tracking
from ..utils import get_anymail_setting, parse_single_address
from .base import AnymailBaseWebhookView, AnymailCoreWebhookView

try:
    # Valid webhook signatures with svix library if available
    from svix.webhooks import Webhook as SvixWebhook, WebhookVerificationError
except ImportError:
    # Otherwise, validating with basic auth is sufficient
    # (unless settings specify signature validation, which will then raise this error)
    SvixWebhook = _LazyError(
        AnymailImproperlyInstalled(missing_package="svix", install_extra="resend")
    )
    WebhookVerificationError = object()


class SvixWebhookValidationMixin(AnymailCoreWebhookView):
    """Mixin to validate Svix webhook signatures"""

    # Consuming classes can override (e.g., to use different secrets
    # for inbound and tracking webhooks).
    _secret_setting_name = "signing_secret"

    @classmethod
    def as_view(cls, **initkwargs):
        if not hasattr(cls, cls._secret_setting_name):
            # The attribute must exist on the class before View.as_view
            # will allow overrides via kwarg
            setattr(cls, cls._secret_setting_name, None)
        return super().as_view(**initkwargs)

    def __init__(self, **kwargs):
        self.signing_secret = get_anymail_setting(
            self._secret_setting_name,
            esp_name=self.esp_name,
            default=None,
            kwargs=kwargs,
        )
        if self.signing_secret is None:
            self._svix_webhook = None
            self.warn_if_no_basic_auth = True
        else:
            # This will raise an import error if svix isn't installed
            self._svix_webhook = SvixWebhook(self.signing_secret)
            # Basic auth is not required if validating signature
            self.warn_if_no_basic_auth = False
        super().__init__(**kwargs)

    def validate_request(self, request):
        if self._svix_webhook:
            # https://docs.svix.com/receiving/verifying-payloads/how
            try:
                # Note: if signature is valid, Svix also tries to parse
                # the json body, so this could raise other errors...
                self._svix_webhook.verify(request.body, request.headers)
            except WebhookVerificationError as error:
                setting_name = f"{self.esp_name}_{self._secret_setting_name}".upper()
                raise AnymailWebhookValidationFailure(
                    f"{self.esp_name} webhook called with incorrect signature"
                    f" (check Anymail {setting_name} setting)"
                ) from error


class ResendTrackingWebhookView(SvixWebhookValidationMixin, AnymailBaseWebhookView):
    """Handler for Resend.com status tracking webhooks"""

    esp_name = "Resend"
    signal = tracking

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))
        return [self.esp_to_anymail_event(esp_event, request)]

    # https://resend.com/docs/dashboard/webhooks/event-types
    event_types = {
        # Map Resend type: Anymail normalized type
        "email.sent": EventType.SENT,
        "email.delivered": EventType.DELIVERED,
        "email.delivery_delayed": EventType.DEFERRED,
        "email.complained": EventType.COMPLAINED,
        "email.bounced": EventType.BOUNCED,
        "email.opened": EventType.OPENED,
        "email.clicked": EventType.CLICKED,
    }

    def esp_to_anymail_event(self, esp_event, request):
        event_type = self.event_types.get(esp_event["type"], EventType.UNKNOWN)

        # event_id: HTTP header `svix-id` is unique for a particular event
        # (including across reposts due to errors)
        try:
            event_id = request.headers["svix-id"]
        except KeyError:
            event_id = None

        # timestamp: Payload created_at is unique for a particular event.
        # (Payload data.created_at is when the message was created, not the event.
        # HTTP header `svix-timestamp` changes for each repost of the same event.)
        try:
            timestamp = datetime.fromisoformat(
                # Must convert "Z" to timezone offset for Python 3.10 and earlier.
                esp_event["created_at"].replace("Z", "+00:00")
            )
        except (KeyError, ValueError):
            timestamp = None

        try:
            message_id = esp_event["data"]["email_id"]
        except (KeyError, TypeError):
            message_id = None

        # Resend doesn't provide bounce reasons or SMTP responses,
        # but it's possible to distinguish some cases by examining
        # the human-readable message text:
        try:
            bounce_message = esp_event["data"]["bounce"]["message"]
        except (KeyError, ValueError):
            bounce_message = None
            reject_reason = None
        else:
            if "suppressed sending" in bounce_message:
                # "Resend has suppressed sending to this address ..."
                reject_reason = RejectReason.BLOCKED
            elif "bounce message" in bounce_message:
                # "The recipient's email provider sent a hard bounce message, ..."
                # "The recipient's email provider sent a general bounce message. ..."
                # "The recipient's email provider sent a bounce message because
                #    the recipient's inbox was full. ..."
                reject_reason = RejectReason.BOUNCED
            else:
                reject_reason = RejectReason.OTHER  # unknown

        # Recover tags and metadata from custom headers
        metadata = {}
        tags = []
        try:
            headers = esp_event["data"]["headers"]
        except KeyError:
            pass
        else:
            for header in headers:
                name = header["name"].lower()
                if name == "x-tags":
                    try:
                        tags = json.loads(header["value"])
                    except (ValueError, TypeError):
                        pass
                elif name == "x-metadata":
                    try:
                        metadata = json.loads(header["value"])
                    except (ValueError, TypeError):
                        pass

        # For multi-recipient emails (including cc and bcc), Resend generates events
        # for each recipient, but no indication of which recipient an event applies to.
        # Just report the first `to` recipient.
        try:
            first_to = esp_event["data"]["to"][0]
            recipient = parse_single_address(first_to).addr_spec
        except (KeyError, IndexError, TypeError, AnymailInvalidAddress):
            recipient = None

        try:
            click_data = esp_event["data"]["click"]
        except (KeyError, TypeError):
            click_url = None
            user_agent = None
        else:
            click_url = click_data.get("link")
            user_agent = click_data.get("userAgent")

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            message_id=message_id,
            event_id=event_id,
            recipient=recipient,
            reject_reason=reject_reason,
            description=bounce_message,
            mta_response=None,
            tags=tags,
            metadata=metadata,
            click_url=click_url,
            user_agent=user_agent,
            esp_event=esp_event,
        )
